%% Load the data    
    galaxyData;
    
    n = size(X,1);
    % rescale inputs between -1 and 1
    minx = min(X);
    maxx = max(X);
    scalefacts = maxx-minx;
    for j=1:size(X,2),
        X(:,j) = 2/scalefacts(j)*(X(:,j)-minx(j))+(-1);
    end
    %rescale outputs between -1 and 1
    y = 2*(y-min(y(:)))/(max(y(:))-min(y(:)))-1;
    
    % take a look at the data
    plot3(X(:,1),X(:,2),y,'.')
    
    %% FIRST WE WILL COMPUTE SIMPLE LINEAR REGRESSION AND ILLUSTRATE
    % COMPUTING Leave-One-Out Cross Validation error
    % Simple linear regression- linear input space
    
    % the solution to the regression weights for a plane through the origin
    % is:
    wlinreg = inv(X'*X)*X'*y; 
    % we can add an offset (variable origin) by augmenting the predictor
    % matrix X
    Xaug = [X ones(n,1)];
    
    wlinreg = inv(Xaug'*Xaug)*Xaug'*y; 
    
     % measure the error
     % we will leave one point out, fit, then compute the error on the left
     % out point.  The average error is the average across all left out
     % points.
     for j = 1:n,
        includedindices = [1:(j-1) (j+1):n];
        wlinreg = inv(Xaug(includedindices,:)'*Xaug(includedindices,:))*Xaug(includedindices,:)'*y(includedindices); 
        ypred = Xaug(j,:)*wlinreg;
        err(j) = (ypred-y(j)).^2;
     end
     meanerr = mean(err);
    
    %% Let's visualize these surfaces
    % First make an x-y grid
    x = -1:0.1:1;
    [X1p,X2p]=meshgrid(x,x);  % grid on a square from -1 to 1 in both x and y
    Xshow = [X1p(:) X2p(:) ones(length(x)^2,1)];
    yshow = Xshow*wlinreg;
    yshowgrid = reshape(yshow,size(X1p));
    
    % do the plot
    surf(X1p,X2p,yshowgrid); hold on;
     plot3(X(:,1),X(:,2),y,'.');
     hold off;
    
   %% QUADRATIC REGRESSION
   % create a new predictor matrix
   Z = [X.^2 X(:,1).*X(:,2) X ones(n,1)];
    for j = 1:n,
        includedindices = [1:(j-1) (j+1):n];
        wlinreg = inv(Z(includedindices,:)'*Z(includedindices,:))*Z(includedindices,:)'*y(includedindices); 
        ypred = Z(j,:)*wlinreg;
        errquad(j) = (ypred-y(j)).^2;
     end
     meanerrquad = mean(errquad);
     
     % note that it doesn't help much, because the surface is fundamentally
     % not quadratic
     
     %% Let's visualize these surfaces
    % First make an x-y grid
    x = -1:0.1:1;
    [X1p,X2p]=meshgrid(x,x);  % grid on a square from -1 to 1 in both x and y
    Zshow = [X1p(:).^2 X2p(:).^2 X1p(:).*X2p(:) X1p(:) X2p(:) ones(length(x)^2,1)];
    yshow = Zshow*wlinreg;
    yshowgrid = reshape(yshow,size(X1p));
    
    % do the plot
    surf(X1p,X2p,yshowgrid); hold on;
     plot3(X(:,1),X(:,2),y,'.');
     hold off;
     
     % NOte that the best fit quadratic is almost linear.
     
     
     %%  RIDGE REGRESSION
     % we need a parameter C that controls the regularization
     C = 0.1;
     I = eye(size(Z,2));
     wridge =  inv(Z'*Z+C*I)*Z'*y; 
     
     
     % LASSO REGRESSION
     % we need a parameter C that controls the trade-off between complexity
     % and error
     
     C = 0.1;
     [w,wp,iteration] = lasso(Z, y, C);
     
     
     % SUPPORT VECTOR REGRESSION
       
    % type help quadprog
    % if you don't have the matlab function "quadprog", compile the c-code
    % qp from Steve Gunn
    % Make sure qp.c, pr_loqo.c, and pr_loqo.h are in your current
    % directory, then try:
    %  If this works replace the quadprog calls below with 
    % qp(...)
    
    mex -v qp.c pr_loqo.c
    
       
    %set up the quadratic programming problem for e-Insensitive Quadratic loss
    % Solve the Optimisation Problem
    % maximize
    % y'*(alphaneg-alphapos)-e*(alphaneg+alphapos)
    %                 -0.5*(alphaneg-alphapos)'*(H+I/C)*(alphaneg-alphapos)
    %                       
    % 
    
    % Sensitivity parameters
     e = 0.02; % insensitive to errors up to e.
     C = 1e2; % controls the trade-off between model complexity and losses - think ridge regression
     % as C -> Inf problem goes to least squares.
    % tolerance for Support Vector Detection
    
    % construct kernel matrix
    Zk = Z*Z';  % this computes the innner products between all feature points
                %  i.e. Zij = 
                % dot( [X1_i^2 X2_i^2 X1_i*X2_i X1_i  X2_i 1],[X1_j^2 X2_j^2 X1_j*X2_j X1_j  X2_j 1])
      
        Zb = [Zk -Zk; -Zk Zk] + eye(2*n)/(2*C);
        c = [(e*ones(n,1) - y); (e*ones(n,1) + y)];  
        vlb = zeros(2*n,1);    % Set the bounds: alphas >= 0
        vub = C*ones(2*n,1);   %                 alphas <= C
        x0 = zeros(2*n,1);
        A = [];
        b = [];
        
   % It helps to make absolutely sure that Zb is not singular (e.g. if C->Inf)
     
   Zb = Zb+1e-10*eye(size(Zb));
 
    
    [alphastar] = quadprog(Zb, c, A, b, [],[],vlb, vub, x0);
    %  OR [alphastar] = qp(Zb, c, A, b, vlb, vub, x0);  this should be
    %  faster
    
    % alphastar is twice the length of the number of data points because we
    % have split the sum of absolute values into two parts.
    % we recover the weight vector by differencing
     beta =  alphastar(1:n) - alphastar(n+1:2*n);
    % our best estimate is 
    ypred = Zk*beta;   
    
    
    % the Problem becomes simpler if we have e = 0.  Then the
    % problem can be rewritten as:
    Zb = Zk + eye(n)/(2*C);
    c = -y;
    Zb = Zb+1e-10*eye(size(Zb))
    [alpha] = quadprog(Zb, c, A, b);
    %[alpha] = qp(Zb, c, A, b);
    
      ypred = Zk*alpha;  
     
     %% RADIAL BASIS FUNCTION REGRESSION

    % Construct the Kernel matrix for RBFs
    sig2 = 0.5;  % sets the kernel width
    % lets visualize this
    circx = cos(2*pi*(0.0:0.05:1));
    circy = sin(2*pi*(0.0:0.05:1));
    for j=1:size(X,1),
        plot(sqrt(2)*sig2*circx+X(j,1),sqrt(2)*sig2*circy+X(j,2)); hold on;
    end
    plot([-1 1 1 -1 -1], [-1 -1 1 1 -1],'k--','LineWidth',2);
    hold off;
    % try sig2 = 0.1 and sig2 = 2 to see the effects of this parameter on
    % the width of the basis functions.  sig2 = 0.5 was selected to match the
    % scaling of the variables (we rescaled to lie between -1 and 1) so
    % that every basis function would have some contribution at each
    % location in the data square.
    
    
    Z = zeros(n,n);  
    for i=1:n
       for j=1:n
          u = X(i,:);
          v = X(j,:);
          Z(i,j) = exp(-(u-v)*(u-v)'/(2*sig2^2));
       end
    end
     
    
    % recall [X1p,X2p]=meshgrid(x,x); from above
    Xp = [X1p(:) X2p(:)];
    
    %MAKE A SECOND MATRIX FOR VISUALIZATION
    % reshape as before to view
    for i=1:size(Xp,1);
        for j=1:n
            u = Xp(i,:);
            v = X(j,:);
            Zshow(i,j) = exp(-(u-v)*(u-v)'/(2*sig2^2));
        end
    end
    
    % As an example visualization, here is a random weight vector
    alpharand = randn(n,1);
    yshow = Zshow*alpharand;  yshow = reshape(yshow,size(X1p));
    surf(X1p,X2p,yshow);
    
    
    
    
    

    
    
    